@@ -1,5 +1,5 @@
 {- 
-    Copyright 2011-2013 Mario Blazevic
+    Copyright 2013-2015 Mario Blazevic
 
     License: BSD3 (see BSD3-LICENSE.txt file)
 -}
@@ -20,26 +20,26 @@
 -- All monoid subclasses listed above are for Abelian, /i.e./, commutative or symmetric monoids. Since most practical
 -- monoids in Haskell are not Abelian, each of the these classes has two symmetric superclasses:
 -- 
--- * 'LeftReductiveMonoid' 
+-- * 'LeftReductiveMonoid'
 -- 
--- * 'LeftCancellativeMonoid' 
+-- * 'LeftCancellativeMonoid'
 -- 
--- * 'LeftGCDMonoid' 
+-- * 'LeftGCDMonoid'
 -- 
--- * 'RightReductiveMonoid' 
+-- * 'RightReductiveMonoid'
 -- 
 -- * 'RightCancellativeMonoid'
 -- 
 -- * 'RightGCDMonoid'
 
-{-# LANGUAGE Haskell2010 #-}
+{-# LANGUAGE Haskell2010, Trustworthy #-}
 
 module Data.Monoid.Cancellative (
    -- * Symmetric, commutative monoid classes
-   CommutativeMonoid, ReductiveMonoid(..), CancellativeMonoid(..), GCDMonoid(..),
+   CommutativeMonoid, ReductiveMonoid(..), CancellativeMonoid, GCDMonoid(..),
    -- * Asymmetric monoid classes
    LeftReductiveMonoid(..), RightReductiveMonoid(..),
-   LeftCancellativeMonoid(..), RightCancellativeMonoid(..),
+   LeftCancellativeMonoid, RightCancellativeMonoid,
    LeftGCDMonoid(..), RightGCDMonoid(..)
    )
 where
@@ -47,7 +47,7 @@ where
 import Prelude hiding (gcd)
 import qualified Prelude
 
-import Data.Monoid (Monoid (mappend), Dual(..), Sum(..), Product(..))
+import Data.Monoid (Monoid, Dual(..), Sum(..), Product(..))
 import qualified Data.List as List
 import Data.Maybe (isJust)
 import qualified Data.ByteString as ByteString
@@ -99,7 +99,7 @@ class (LeftCancellativeMonoid m, RightCancellativeMonoid m, ReductiveMonoid m) =
 class (ReductiveMonoid m, LeftGCDMonoid m, RightGCDMonoid m) => GCDMonoid m where
    gcd :: m -> m -> m
 
--- | Class of monoids with a left inverse of 'mappend', satisfying the following law:
+-- | Class of monoids with a left inverse of 'Data.Monoid.mappend', satisfying the following law:
 -- 
 -- > isPrefixOf a b == isJust (stripPrefix a b)
 -- > maybe b (a <>) (stripPrefix a b) == b
@@ -113,7 +113,7 @@ class Monoid m => LeftReductiveMonoid m where
 
    isPrefixOf a b = isJust (stripPrefix a b)
 
--- | Class of monoids with a right inverse of 'mappend', satisfying the following law:
+-- | Class of monoids with a right inverse of 'Data.Monoid.mappend', satisfying the following law:
 -- 
 -- > isSuffixOf a b == isJust (stripSuffix a b)
 -- > maybe b (<> a) (stripSuffix a b) == b
@@ -273,7 +273,7 @@ instance Num a => CommutativeMonoid (Product a)
 
 instance Integral a => ReductiveMonoid (Product a) where
    Product 0 </> Product 0 = Just (Product 0)
-   Product a </> Product 0 = Nothing
+   Product _ </> Product 0 = Nothing
    Product a </> Product b = if remainder == 0 then Just (Product quotient) else Nothing
       where (quotient, remainder) = quotRem a b
 
@@ -414,7 +414,7 @@ instance Ord k => LeftReductiveMonoid (Map.Map k a) where
                    | otherwise = Nothing
 
 instance (Ord k, Eq a) => LeftGCDMonoid (Map.Map k a) where
-   commonPrefix = Map.mergeWithKey (\k a b -> if a == b then Just a else Nothing) (const Map.empty) (const Map.empty)
+   commonPrefix = Map.mergeWithKey (\_ a b -> if a == b then Just a else Nothing) (const Map.empty) (const Map.empty)
 
 -- IntMap instances
 
@@ -424,7 +424,7 @@ instance LeftReductiveMonoid (IntMap.IntMap a) where
                    | otherwise = Nothing
 
 instance Eq a => LeftGCDMonoid (IntMap.IntMap a) where
-   commonPrefix = IntMap.mergeWithKey (\k a b -> if a == b then Just a else Nothing)
+   commonPrefix = IntMap.mergeWithKey (\_ a b -> if a == b then Just a else Nothing)
                                       (const IntMap.empty) (const IntMap.empty)
 
 -- List instances
@@ -439,7 +439,7 @@ instance Eq x => LeftGCDMonoid [x] where
    commonPrefix (x:xs) (y:ys) | x == y = x : commonPrefix xs ys
    commonPrefix _ _ = []
 
-   stripCommonPrefix x y = strip' id x y
+   stripCommonPrefix x0 y0 = strip' id x0 y0
       where strip' f (x:xs) (y:ys) | x == y = strip' (f . (x :)) xs ys
             strip' f x y = (f [], x, y)
 
@@ -542,8 +542,8 @@ instance RightCancellativeMonoid ByteString.ByteString
 instance LeftGCDMonoid ByteString.ByteString where
    stripCommonPrefix x y = (xp, xs, ByteString.unsafeDrop maxPrefixLength y)
       where maxPrefixLength = prefixLength 0 (ByteString.length x `min` ByteString.length y)
-            prefixLength n len | n < len, 
-                                 ByteString.unsafeIndex x n == ByteString.unsafeIndex y n = 
+            prefixLength n len | n < len,
+                                 ByteString.unsafeIndex x n == ByteString.unsafeIndex y n =
                                     prefixLength (succ n) len
                                | otherwise = n
             (xp, xs) = ByteString.splitAt maxPrefixLength x
@@ -577,7 +577,7 @@ instance RightCancellativeMonoid LazyByteString.ByteString
 instance LeftGCDMonoid LazyByteString.ByteString where
    stripCommonPrefix x y = (xp, xs, LazyByteString.drop maxPrefixLength y)
       where maxPrefixLength = prefixLength 0 (LazyByteString.length x `min` LazyByteString.length y)
-            prefixLength n len | n < len && LazyByteString.index x n == LazyByteString.index y n = 
+            prefixLength n len | n < len && LazyByteString.index x n == LazyByteString.index y n =
                prefixLength (succ n) len
             prefixLength n _ = n
             (xp, xs) = LazyByteString.splitAt maxPrefixLength x
