@@ -1,5 +1,5 @@
 {- 
-    Copyright 2011-2013 Mario Blazevic
+    Copyright 2013-2015 Mario Blazevic
 
     License: BSD3 (see BSD3-LICENSE.txt file)
 -}
@@ -10,17 +10,14 @@
 {-# LANGUAGE Haskell2010 #-}
 
 module Data.Monoid.Instances.Concat (
-   Concat, inject, extract 
+   Concat, concatenate, extract
    )
 where
 
-import Prelude hiding (all, any, break, filter, foldl, foldl1, foldr, foldr1, map, concatMap, 
+import Prelude hiding (all, any, break, filter, foldl, foldl1, foldMap, foldr, foldr1, map, concatMap,
                        length, null, reverse, scanl, scanr, scanl1, scanr1, span, splitAt)
-import Data.Foldable (Foldable)
-import Data.Traversable (Traversable, traverse)
+import Control.Applicative (Applicative(..))
 import qualified Data.Foldable as Foldable
-import qualified Data.Traversable as Traversable
-import Data.Maybe (fromMaybe)
 import Data.String (IsString(..))
 import Data.Monoid (Monoid(..), (<>), First(..), Sum(..))
 import Data.Monoid.Cancellative (LeftReductiveMonoid(..), RightReductiveMonoid(..),
@@ -34,19 +31,30 @@ import Data.Sequence (Seq, empty, filter, (<|), (|>), ViewL((:<)), ViewR((:>)))
 import qualified Data.Sequence as Seq
 
 -- | @'Concat' a@ is a @newtype@ wrapper around @'Seq' a@. The behaviour of the @'Concat' a@ instances of monoid
--- subclasses is identical to the behaviour of their @a@ instances, up to the @'inject' . 'singleton'@ isomorphism.
+-- subclasses is identical to the behaviour of their @a@ instances, up to the 'pure' isomorphism.
 --
 -- The only purpose of 'Concat' then is to change the performance characteristics of various operations. Most
 -- importantly, injecting a monoid into a 'Concat' has the effect of making 'mappend' a constant-time operation.
 --
 newtype Concat a = Concat {extract :: Seq a} deriving Show
 
+concatenate :: (MonoidNull a, PositiveMonoid a) => Seq a -> Concat a
+concatenate = Concat . filter (not . null)
+
 instance (Eq a, Monoid a) => Eq (Concat a) where
    Concat x == Concat y = Foldable.foldMap id x == Foldable.foldMap id y
 
 instance (Ord a, Monoid a) => Ord (Concat a) where
    compare (Concat x) (Concat y) = compare (Foldable.foldMap id x) (Foldable.foldMap id y)
 
+instance Functor Concat where
+   fmap f (Concat x) = Concat (fmap f x)
+
+instance Applicative Concat where
+   pure a = Concat (Seq.singleton a)
+   Concat x <*> Concat y = Concat (x <*> y)
+   Concat x *> Concat y = Concat (x *> y)
+
 instance Monoid (Concat a) where
    mempty = Concat Seq.empty
    mappend (Concat a) (Concat b) = Concat (mappend a b)
@@ -57,7 +65,7 @@ instance MonoidNull (Concat a) where
 instance PositiveMonoid (Concat a)
 
 instance (LeftReductiveMonoid a, MonoidNull a, StableFactorialMonoid a) => LeftReductiveMonoid (Concat a) where
-   stripPrefix (Concat x) (Concat y) = fmap Concat $ strip1 x y
+   stripPrefix c1 c2 = fmap Concat $ strip1 (extract c1) (extract c2)
       where strip1 x y = strip2 (Seq.viewl x) y
             strip2 Seq.EmptyL y = Just y
             strip2 (xp :< xs) y = strip3 xp xs (Seq.viewl y)
@@ -69,7 +77,7 @@ instance (LeftReductiveMonoid a, MonoidNull a, StableFactorialMonoid a) => LeftR
                   (Nothing, Just xps) -> strip3 xps xs (Seq.viewl ys)
 
 instance (MonoidNull a, RightReductiveMonoid a, StableFactorialMonoid a) => RightReductiveMonoid (Concat a) where
-   stripSuffix (Concat x) (Concat y) = fmap Concat $ strip1 x y
+   stripSuffix c1 c2 = fmap Concat $ strip1 (extract c1) (extract c2)
       where strip1 x y = strip2 (Seq.viewr x) y
             strip2 Seq.EmptyR y = Just y
             strip2 (xp :> xs) y = strip3 xp xs (Seq.viewr y)
@@ -130,11 +138,11 @@ instance FactorialMonoid a => FactorialMonoid (Concat a) where
               xp :> xs -> Just (Concat xp', Concat $ Seq.singleton xss)
                  where Just (xsp, xss) = splitPrimeSuffix xs
                        xp' = if null xsp then xp else xp |> xsp
-   foldl f a (Concat x) = Foldable.foldl g a x
+   foldl f a0 (Concat x) = Foldable.foldl g a0 x
       where g = Factorial.foldl (\a-> f a . Concat . Seq.singleton)
-   foldl' f a (Concat x) = Foldable.foldl' g a x
+   foldl' f a0 (Concat x) = Foldable.foldl' g a0 x
       where g = Factorial.foldl' (\a-> f a . Concat . Seq.singleton)
-   foldr f a (Concat x) = Foldable.foldr g a x
+   foldr f a0 (Concat x) = Foldable.foldr g a0 x
       where g a b = Factorial.foldr (f . Concat . Seq.singleton) b a
    length (Concat x) = getSum $ Foldable.foldMap (Sum . length) x
    foldMap f (Concat x) = Foldable.foldMap (foldMap (f . Concat . Seq.singleton)) x
@@ -147,7 +155,7 @@ instance FactorialMonoid a => FactorialMonoid (Concat a) where
             where (xpp, xps) = Factorial.span (p . Concat . Seq.singleton) xp
                   (Concat xsp, xss) = Factorial.span p (Concat xs)
    split p (Concat x) = Foldable.foldr splitNext [mempty] x
-      where splitNext a (xp:xs) =
+      where splitNext a ~(xp:xs) =
                let as = fmap (Concat . Seq.singleton) (Factorial.split (p . Concat . Seq.singleton) a)
                in if null xp
                   then as ++ xs
@@ -159,7 +167,7 @@ instance FactorialMonoid a => FactorialMonoid (Concat a) where
          xp :< xs | k < n -> (Concat (xp <| xsp), xss)
                   | otherwise -> (Concat $ Seq.singleton xpp, Concat (if null xps then xs else xps <| xs))
             where k = length xp
-                  (Concat xsp, xss) = splitAt (n - k) (Concat xs) 
+                  (Concat xsp, xss) = splitAt (n - k) (Concat xs)
                   (xpp, xps) = splitAt n xp
    reverse (Concat x) = Concat (fmap reverse $ reverse x)
 
@@ -186,11 +194,11 @@ instance (Eq a, TextualMonoid a, StableFactorialMonoid a) => TextualMonoid (Conc
    any p (Concat x) = Foldable.any (any p) x
    all p (Concat x) = Foldable.all (all p) x
 
-   foldl ft fc a (Concat x) = Foldable.foldl g a x
+   foldl ft fc a0 (Concat x) = Foldable.foldl g a0 x
       where g = Textual.foldl (\a-> ft a . Concat . Seq.singleton) fc
-   foldl' ft fc a (Concat x) = Foldable.foldl' g a x
+   foldl' ft fc a0 (Concat x) = Foldable.foldl' g a0 x
       where g = Textual.foldl' (\a-> ft a . Concat . Seq.singleton) fc
-   foldr ft fc a (Concat x) = Foldable.foldr g a x
+   foldr ft fc a0 (Concat x) = Foldable.foldr g a0 x
       where g a b = Textual.foldr (ft . Concat . Seq.singleton) fc b a
 
    span pt pc (Concat x) =
@@ -204,10 +212,3 @@ instance (Eq a, TextualMonoid a, StableFactorialMonoid a) => TextualMonoid (Conc
    break pt pc = Textual.span (not . pt) (not . pc)
 
    find p (Concat x) = getFirst $ Foldable.foldMap (First . find p) x
-
-inject :: (MonoidNull a, PositiveMonoid a) => Seq a -> Concat a
-inject = Concat . filter (not . null)
-
-injectSingleton :: (MonoidNull a, PositiveMonoid a) => a -> Concat a
-injectSingleton a | null a = mempty
-                  | otherwise = Concat (Seq.singleton a)
