@@ -1,5 +1,5 @@
 {- 
-    Copyright 2013 Mario Blazevic
+    Copyright 2013-2015 Mario Blazevic
 
     License: BSD3 (see BSD3-LICENSE.txt file)
 -}
@@ -11,8 +11,9 @@ module Main where
 
 import Prelude hiding (foldl, foldr, gcd, length, null, reverse, span, splitAt, takeWhile)
 
-import Test.QuickCheck (Arbitrary, CoArbitrary, Property, Gen,
-                        quickCheck, arbitrary, coarbitrary, property, label, forAll, variant, whenFail, (.&&.))
+import Test.Tasty (defaultMain, testGroup)
+import Test.Tasty.QuickCheck (Arbitrary, CoArbitrary, Property, Gen,
+                              arbitrary, coarbitrary, property, label, forAll, mapSize, testProperty, variant, whenFail, (.&&.))
 import Test.QuickCheck.Instances ()
 
 import Control.Applicative (Applicative(..), liftA2)
@@ -51,13 +52,15 @@ import Data.Monoid.Instances.Measured (Measured)
 import qualified Data.Monoid.Instances.Measured as Measured
 import Data.Monoid.Instances.Stateful (Stateful)
 import qualified Data.Monoid.Instances.Stateful as Stateful
+import Data.Monoid.Instances.Positioned (OffsetPositioned, LinePositioned)
+import qualified Data.Monoid.Instances.Positioned as Positioned
 
 import Data.Monoid (Monoid, mempty, (<>), mconcat, All(All), Any(Any), Dual(Dual),
                     First(First), Last(Last), Sum(Sum), Product(Product))
 import Data.Monoid.Null (MonoidNull, PositiveMonoid, null)
 import Data.Monoid.Factorial (FactorialMonoid, StableFactorialMonoid, 
-                              factors, splitPrimePrefix, splitPrimeSuffix, primePrefix, primeSuffix,
-                              foldl, foldl', foldr, length, reverse, span, split, splitAt)
+                              factors, splitPrimePrefix, splitPrimeSuffix, primePrefix, primeSuffix, inits, tails,
+                              foldl, foldl', foldr, length, reverse, span, spanMaybe, split, splitAt)
 import Data.Monoid.Cancellative (CommutativeMonoid, ReductiveMonoid, LeftReductiveMonoid, RightReductiveMonoid,
                                  CancellativeMonoid, LeftCancellativeMonoid, RightCancellativeMonoid,
                                  GCDMonoid, LeftGCDMonoid, RightGCDMonoid,
@@ -173,7 +176,7 @@ factorialInstances = map upcast stableFactorialInstances
    where upcast (StableFactorialMonoidInstance i) = FactorialMonoidInstance i
 
 stableFactorialInstances :: [StableFactorialMonoidInstance]
-stableFactorialInstances = stable1 ++ map measure stable1
+stableFactorialInstances = stable1 ++ map measure stable1 ++ map position stable1 
    where stable1 = map upcast stableTextualInstances
                    ++ [StableFactorialMonoidInstance (mempty :: ByteString),
                        StableFactorialMonoidInstance (mempty :: Lazy.ByteString),
@@ -181,7 +184,9 @@ stableFactorialInstances = stable1 ++ map measure stable1
                        StableFactorialMonoidInstance (mempty :: Seq Int),
                        StableFactorialMonoidInstance (mempty :: Vector Int)]
          upcast (StableTextualMonoidInstance i) = StableFactorialMonoidInstance i
-         measure (StableFactorialMonoidInstance i) = StableFactorialMonoidInstance (Measured.inject i)
+         measure (StableFactorialMonoidInstance i) = StableFactorialMonoidInstance (Measured.measure i)
+         position (StableFactorialMonoidInstance (i :: a)) = 
+            StableFactorialMonoidInstance (pure i :: OffsetPositioned a)
 
 textualInstances :: [TextualMonoidInstance]
 textualInstances = map upcast stableTextualInstances
@@ -194,12 +199,17 @@ textualInstances = map upcast stableTextualInstances
    where upcast (StableTextualMonoidInstance i) = TextualMonoidInstance i
 
 stableTextualInstances :: [StableTextualMonoidInstance]
-stableTextualInstances = [StableTextualMonoidInstance (mempty :: TestString),
-                          StableTextualMonoidInstance (mempty :: String),
-                          StableTextualMonoidInstance (mempty :: Text),
-                          StableTextualMonoidInstance (mempty :: Lazy.Text),
-                          StableTextualMonoidInstance (mempty :: Seq Char),
-                          StableTextualMonoidInstance (mempty :: Vector Char)]
+stableTextualInstances = stable1 ++ map measure stable1 ++ concatMap position stable1
+   where stable1 = [StableTextualMonoidInstance (mempty :: TestString),
+                    StableTextualMonoidInstance (mempty :: String),
+                    StableTextualMonoidInstance (mempty :: Text),
+                    StableTextualMonoidInstance (mempty :: Lazy.Text),
+                    StableTextualMonoidInstance (mempty :: Seq Char),
+                    StableTextualMonoidInstance (mempty :: Vector Char)]
+         measure (StableTextualMonoidInstance i) = StableTextualMonoidInstance (Measured.measure i)
+         position (StableTextualMonoidInstance (i :: a)) = 
+            [StableTextualMonoidInstance (pure i :: OffsetPositioned a),
+             StableTextualMonoidInstance (pure i :: LinePositioned a)]
 
 leftReductiveInstances = map upcast leftCancellativeInstances
                          ++ [LeftReductiveMonoidInstance (mempty :: Sum Integer),
@@ -301,29 +311,26 @@ cancellativeGCDInstances = [CancellativeGCDMonoidInstance (),
                             CancellativeGCDMonoidInstance (mempty :: Dual (Sum Integer)),
                             CancellativeGCDMonoidInstance (mempty :: (Sum Integer, Sum Int))]
 
-main = mapM_ (quickCheck . uncurry checkInstances) tests
+main = defaultMain (testGroup "MonoidSubclasses" $ map expand tests)
+  where expand (name, test) = testProperty name (foldr1 (.&&.) $ checkInstances test)
 
-checkInstances :: String -> Test -> Property
-checkInstances name (CommutativeTest checkType) = label name $ foldr1 (.&&.) (map checkType commutativeInstances)
-checkInstances name (NullTest checkType) = label name $ foldr1 (.&&.) (map checkType nullInstances)
-checkInstances name (PositiveTest checkType) = label name $ foldr1 (.&&.) (map checkType positiveInstances)
-checkInstances name (FactorialTest checkType) = label name $ foldr1 (.&&.) (map checkType factorialInstances)
-checkInstances name (StableFactorialTest checkType) =
-   label name $ foldr1 (.&&.) (map checkType stableFactorialInstances)
-checkInstances name (TextualTest checkType) = label name $ foldr1 (.&&.) (map checkType textualInstances)
-checkInstances name (LeftReductiveTest checkType) = label name $ foldr1 (.&&.) (map checkType leftReductiveInstances)
-checkInstances name (RightReductiveTest checkType) = label name $ foldr1 (.&&.) (map checkType rightReductiveInstances)
-checkInstances name (ReductiveTest checkType) = label name $ foldr1 (.&&.) (map checkType reductiveInstances)
-checkInstances name (LeftCancellativeTest checkType) =
-   label name $ foldr1 (.&&.) (map checkType leftCancellativeInstances) 
-checkInstances name (RightCancellativeTest checkType) =
-   label name $ foldr1 (.&&.) (map checkType rightCancellativeInstances) 
-checkInstances name (CancellativeTest checkType) = label name $ foldr1 (.&&.) (map checkType cancellativeInstances) 
-checkInstances name (LeftGCDTest checkType) = label name $ foldr1 (.&&.) (map checkType leftGCDInstances) 
-checkInstances name (RightGCDTest checkType) = label name $ foldr1 (.&&.) (map checkType rightGCDInstances) 
-checkInstances name (GCDTest checkType) = label name $ foldr1 (.&&.) (map checkType gcdInstances)  
-checkInstances name (CancellativeGCDTest checkType) = 
-   label name $ foldr1 (.&&.) (map checkType cancellativeGCDInstances) 
+checkInstances :: Test -> [Property]
+checkInstances (CommutativeTest checkType) = (map checkType commutativeInstances)
+checkInstances (NullTest checkType) = (map checkType nullInstances)
+checkInstances (PositiveTest checkType) = (map checkType positiveInstances)
+checkInstances (FactorialTest checkType) = (map checkType factorialInstances)
+checkInstances (StableFactorialTest checkType) = (map checkType stableFactorialInstances)
+checkInstances (TextualTest checkType) = (map checkType textualInstances)
+checkInstances (LeftReductiveTest checkType) = (map checkType leftReductiveInstances)
+checkInstances (RightReductiveTest checkType) = (map checkType rightReductiveInstances)
+checkInstances (ReductiveTest checkType) = (map checkType reductiveInstances)
+checkInstances (LeftCancellativeTest checkType) = (map checkType leftCancellativeInstances) 
+checkInstances (RightCancellativeTest checkType) = (map checkType rightCancellativeInstances) 
+checkInstances (CancellativeTest checkType) = (map checkType cancellativeInstances) 
+checkInstances (LeftGCDTest checkType) = (map checkType leftGCDInstances) 
+checkInstances (RightGCDTest checkType) = (map checkType rightGCDInstances) 
+checkInstances (GCDTest checkType) = (map checkType gcdInstances)  
+checkInstances (CancellativeGCDTest checkType) = (map checkType cancellativeGCDInstances) 
 
 tests :: [(String, Test)]
 tests = [("CommutativeMonoid", CommutativeTest checkCommutative),
@@ -335,11 +342,14 @@ tests = [("CommutativeMonoid", CommutativeTest checkCommutative),
          ("splitPrimeSuffix", FactorialTest checkSplitPrimeSuffix),
          ("primePrefix", FactorialTest checkPrimePrefix),
          ("primeSuffix", FactorialTest checkPrimeSuffix),
+         ("inits", FactorialTest checkInits),
+         ("tails", FactorialTest checkTails),
          ("foldl", FactorialTest checkLeftFold),
          ("foldl'", FactorialTest checkLeftFold'),
          ("foldr", FactorialTest checkRightFold),
          ("length", FactorialTest checkLength),
          ("span", FactorialTest checkSpan),
+         ("spanMaybe", FactorialTest checkSpanMaybe),
          ("split", FactorialTest checkSplit),
          ("splitAt", FactorialTest checkSplitAt),
          ("reverse", FactorialTest checkReverse),
@@ -368,8 +378,18 @@ tests = [("CommutativeMonoid", CommutativeTest checkCommutative),
          ("Textual.dropWhile", TextualTest checkTextualDropWhile),
          ("Textual.span", TextualTest checkTextualSpan),
          ("Textual.break", TextualTest checkTextualBreak),
+         ("Textual.spanMaybe", TextualTest checkTextualSpanMaybe),
          ("Textual.split", TextualTest checkTextualSplit),
          ("Textual.find", TextualTest checkTextualFind),
+         ("Textual.foldl_", TextualTest checkTextualFoldl_),
+         ("Textual.foldr_", TextualTest checkTextualFoldr_),
+         ("Textual.foldl_'", TextualTest checkTextualFoldl_'),
+         ("Textual.span_", TextualTest checkTextualSpan_),
+         ("Textual.break_", TextualTest checkTextualBreak_),
+         ("Textual.spanMaybe_", TextualTest checkTextualSpanMaybe_),
+         ("Textual.spanMaybe_'", TextualTest checkTextualSpanMaybe_'),
+         ("Textual.takeWhile_", TextualTest checkTextualTakeWhile_),
+         ("Textual.dropWhile_", TextualTest checkTextualDropWhile_),
          ("stripPrefix", LeftReductiveTest checkStripPrefix),
          ("isPrefixOf", LeftReductiveTest checkIsPrefixOf),
          ("stripSuffix", RightReductiveTest checkStripSuffix),
@@ -411,6 +431,12 @@ checkPrimePrefix (FactorialMonoidInstance (_ :: a)) =
 checkPrimeSuffix (FactorialMonoidInstance (_ :: a)) = 
    forAll (arbitrary :: Gen a) (\a-> primeSuffix a == maybe mempty snd (splitPrimeSuffix a))
 
+checkInits (FactorialMonoidInstance (_ :: a)) =
+   mapSize (`div` 5) $ forAll (arbitrary :: Gen a) (\a-> inits a == List.map mconcat (List.inits $ factors a))
+
+checkTails (FactorialMonoidInstance (_ :: a)) =
+   mapSize (`div` 5) $ forAll (arbitrary :: Gen a) (\a-> tails a == List.map mconcat (List.tails $ factors a))
+
 checkLeftFold (FactorialMonoidInstance (_ :: a)) = 
    forAll (arbitrary :: Gen a) (\a-> foldl (flip (:)) [] a == List.foldl (flip (:)) [] (factors a))
 
@@ -427,6 +453,14 @@ checkSpan (FactorialMonoidInstance (_ :: a)) = property $ \p-> forAll (arbitrary
    where check p a = span p a == (mconcat l, mconcat r)
             where (l, r) = List.span p (factors a)
 
+checkSpanMaybe (FactorialMonoidInstance (_ :: a)) = property $ \(f, s)-> forAll (arbitrary :: Gen a) (check f (s :: Bool))
+   where check f s0 a = a == prefix <> suffix
+                        && foldMaybe prefix == Just s'
+                        && (null suffix || f s' (primePrefix suffix) == Nothing)
+            where (prefix, suffix, s') = spanMaybe s0 f a
+                  foldMaybe = foldl g (Just s0)
+                  g s m = s >>= flip f m
+
 checkSplit (FactorialMonoidInstance (_ :: a)) = forAll (arbitrary :: Gen a) check
    where check a = property (\pred-> all (all (not . pred) . factors) (split pred a))
                    .&&. property (\prime-> mconcat (intersperse prime $ split (== prime) a) == a)
@@ -495,7 +529,7 @@ checkTextualFoldr (TextualMonoidInstance (_ :: a)) =
    forAll (arbitrary :: Gen a) check1 .&&. forAll (arbitrary :: Gen String) check2
    where check1 a = Textual.foldr (\a l-> Left a : l) (\c l-> Right c : l) [] a == textualFactors a
                     && Textual.foldr (<>) ((<>) . Textual.singleton) mempty a == a
-         check2 s = Textual.foldr undefined (:) [] s == s
+         check2 s = Textual.foldr undefined (:) [] (fromString s :: a) == s
 
 checkTextualFoldl' (TextualMonoidInstance (_ :: a)) = 
    forAll (arbitrary :: Gen a) check1 .&&. forAll (arbitrary :: Gen String) check2
@@ -503,7 +537,22 @@ checkTextualFoldl' (TextualMonoidInstance (_ :: a)) =
                     && Textual.foldl' (<>) (\a-> (a <>) . Textual.singleton) mempty a == a
          check2 s = Textual.foldl' undefined (flip (:)) [] s == List.foldl' (flip (:)) [] s
 
-checkTextualScanl (TextualMonoidInstance (_ :: a)) = 
+checkTextualFoldl_ (TextualMonoidInstance (_ :: a)) =
+   forAll (arbitrary :: Gen a) check1 .&&. forAll (arbitrary :: Gen String) check2
+   where check1 a = Textual.foldl_ (\l c-> c : l) [] a == List.reverse (rights $ textualFactors a)
+         check2 s = Textual.foldl_ (flip (:)) [] s == List.foldl (flip (:)) [] s
+
+checkTextualFoldr_ (TextualMonoidInstance (_ :: a)) =
+   forAll (arbitrary :: Gen a) check1 .&&. forAll (arbitrary :: Gen String) check2
+   where check1 a = Textual.foldr_ (\c l-> c : l) [] a == rights (textualFactors a)
+         check2 s = Textual.foldr_ (:) [] (fromString s :: a) == s
+
+checkTextualFoldl_' (TextualMonoidInstance (_ :: a)) =
+   forAll (arbitrary :: Gen a) check1 .&&. forAll (arbitrary :: Gen String) check2
+   where check1 a = Textual.foldl_' (\l c-> c : l) [] a == List.reverse (rights $ textualFactors a)
+         check2 s = Textual.foldl_' (flip (:)) [] s == List.foldl (flip (:)) [] s
+
+checkTextualScanl (TextualMonoidInstance (_ :: a)) =
    forAll (arbitrary :: Gen a) check1 .&&. forAll (arbitrary :: Gen String) check2
    where check1 a = (rights . textualFactors . Textual.scanl f 'Z') a == (List.scanl f 'Z' . rights . textualFactors) a
                     && (lefts . textualFactors . Textual.scanl f 'Y') a == (lefts . textualFactors) a
@@ -511,7 +560,7 @@ checkTextualScanl (TextualMonoidInstance (_ :: a)) =
          check2 s = Textual.scanl f 'X' (fromString s :: a) == fromString (List.scanl f 'X' s)
          f c1 c2 = min c1 c2
 
-checkTextualScanr (TextualMonoidInstance (_ :: a)) = 
+checkTextualScanr (TextualMonoidInstance (_ :: a)) =
    forAll (arbitrary :: Gen a) check1 .&&. forAll (arbitrary :: Gen String) check2
    where check1 a = (rights . textualFactors . Textual.scanr f 'Z') a == (List.scanr f 'Z' . rights . textualFactors) a
                     && (lefts . textualFactors . Textual.scanr f 'Y') a == (lefts . textualFactors) a
@@ -519,7 +568,7 @@ checkTextualScanr (TextualMonoidInstance (_ :: a)) =
          check2 s = Textual.scanr f 'X' (fromString s :: a) == fromString (List.scanr f 'X' s)
          f c1 c2 = min c1 c2
 
-checkTextualScanl1 (TextualMonoidInstance (_ :: a)) = 
+checkTextualScanl1 (TextualMonoidInstance (_ :: a)) =
    forAll (arbitrary :: Gen a) check1 .&&. forAll (arbitrary :: Gen String) check2
    where check1 a = Textual.scanl1 (const id) a == a
          check2 s = Textual.scanl1 f (fromString s :: a) == fromString (List.scanl1 f s)
@@ -569,6 +618,56 @@ checkTextualBreak (TextualMonoidInstance (_ :: a)) = forAll (arbitrary :: Gen a)
             where pt = (/= primePrefix a)
          pc = isLetter
 
+checkTextualSpanMaybe (TextualMonoidInstance (_ :: a)) =
+   property $ \(ft, fc, s)-> forAll (arbitrary :: Gen a) (check ft fc (s :: Bool))
+   where check ft fc s0 a = a == prefix <> suffix
+                            && foldMaybe prefix == Just s'
+                            && (null suffix
+                                || maybe (ft s' (primePrefix suffix)) (fc s') (Textual.characterPrefix suffix) == Nothing)
+            where (prefix, suffix, s') = Textual.spanMaybe s0 ft fc a
+                  foldMaybe = Textual.foldl gt gc (Just s0)
+                  gt s m = s >>= flip ft m
+                  gc s c = s >>= flip fc c
+
+checkTextualSpan_ (TextualMonoidInstance (_ :: a)) = forAll (arbitrary :: Gen (a, Bool)) check
+   where check (a, bt) = Textual.span_ bt isLetter a == (Textual.takeWhile_ bt isLetter a, Textual.dropWhile_ bt isLetter a)
+
+checkTextualBreak_ (TextualMonoidInstance (_ :: a)) = forAll (arbitrary :: Gen (a, Bool)) check
+   where check (a, bt) = Textual.break_ bt isLetter a == Textual.span_ (not bt) (not . isLetter) a
+
+checkTextualSpanMaybe_ (TextualMonoidInstance (_ :: a)) =
+   property $ \(fc, s)-> forAll (arbitrary :: Gen a) (check fc (s :: Bool))
+   where check fc s0 a = a == prefix <> suffix
+                         && foldMaybe prefix == Just s'
+                         && (null suffix || (Textual.characterPrefix suffix >>= fc s') == Nothing)
+            where (prefix, suffix, s') = Textual.spanMaybe_ s0 fc a
+                  foldMaybe = Textual.foldl_ gc (Just s0)
+                  gc s c = s >>= flip fc c
+
+checkTextualSpanMaybe_' (TextualMonoidInstance (_ :: a)) =
+   property $ \(fc, s)-> forAll (arbitrary :: Gen a) (check fc (s :: Bool))
+   where check fc s0 a = a == prefix <> suffix
+                         && foldMaybe prefix == Just s'
+                         && (null suffix || (Textual.characterPrefix suffix >>= fc s') == Nothing)
+            where (prefix, suffix, s') = Textual.spanMaybe_' s0 fc a
+                  foldMaybe = Textual.foldl_' gc (Just s0)
+                  gc s c = s >>= flip fc c
+
+checkTextualTakeWhile_ (TextualMonoidInstance (_ :: a)) = 
+   forAll (arbitrary :: Gen a) check1 .&&. forAll (arbitrary :: Gen String) check2
+   where check1 a = textualFactors (Textual.takeWhile_ True isLetter a)
+                    == List.takeWhile (either (const True) isLetter) (textualFactors a)
+                    && Textual.takeWhile_ True (const True) a == a
+         check2 s = Textual.takeWhile_ undefined isLetter (fromString s :: a) == fromString (List.takeWhile isLetter s)
+
+checkTextualDropWhile_ (TextualMonoidInstance (_ :: a)) = 
+   forAll (arbitrary :: Gen a) check1 .&&. forAll (arbitrary :: Gen String) check2
+   where check1 a = textualFactors (Textual.dropWhile_ True isLetter a)
+                    == List.dropWhile (either (const True) isLetter) (textualFactors a)
+                    && Textual.dropWhile_ False (const False) a == a
+         check2 s = Textual.dropWhile_ undefined isLetter (fromString s :: a)
+                    == fromString (List.dropWhile isLetter s)
+
 checkTextualSplit (TextualMonoidInstance (_ :: a)) = forAll (arbitrary :: Gen a) check
    where check a = List.all (List.all isLetter . rights . textualFactors) (Textual.split (not . isLetter) a)
                    && (mconcat . intersperse (fromString " ") . Textual.split (== ' ')) a == a
@@ -683,10 +782,16 @@ instance Arbitrary ByteStringUTF8 where
    arbitrary = fmap ByteStringUTF8 arbitrary
 
 instance (Arbitrary a, MonoidNull a, PositiveMonoid a) => Arbitrary (Concat a) where
-   arbitrary = fmap Concat.inject arbitrary
+   arbitrary = fmap Concat.concatenate arbitrary
 
 instance (Arbitrary a, FactorialMonoid a) => Arbitrary (Measured a) where
-   arbitrary = fmap Measured.inject arbitrary
+   arbitrary = fmap Measured.measure arbitrary
+
+instance (Arbitrary a, FactorialMonoid a) => Arbitrary (OffsetPositioned a) where
+   arbitrary = fmap pure arbitrary
+
+instance (Arbitrary a, TextualMonoid a) => Arbitrary (LinePositioned a) where
+   arbitrary = fmap pure arbitrary
 
 instance (Arbitrary a, Arbitrary b) => Arbitrary (Stateful a b) where
    arbitrary = Stateful.Stateful <$> liftA2 (,) arbitrary arbitrary
@@ -724,12 +829,21 @@ instance CoArbitrary a => CoArbitrary (Concat a) where
 instance CoArbitrary a => CoArbitrary (Measured a) where
    coarbitrary = coarbitrary . Measured.extract
 
+instance CoArbitrary a => CoArbitrary (OffsetPositioned a) where
+   coarbitrary = coarbitrary . Positioned.extract
+
+instance CoArbitrary a => CoArbitrary (LinePositioned a) where
+   coarbitrary = coarbitrary . Positioned.extract
+
 instance CoArbitrary b => CoArbitrary (Stateful a b) where
    coarbitrary = coarbitrary . Stateful.extract
 
 instance Show a => Show (a -> Bool) where
    show _ = "predicate"
 
+instance Show a => Show (Bool -> a -> Maybe Bool) where
+   show _ = "stateful predicate"
+
 instance (PositiveMonoid a, MonoidNull b) => PositiveMonoid (a, b)
 
 #if MIN_VERSION_containers(0,5,2)
@@ -739,4 +853,4 @@ instance Applicative Seq where
    fs <*> xs = Foldable.foldl' add mempty fs
       where add ys f = ys <> fmap f xs
 
-#endif
\ No newline at end of file
+#endif
